#    Copyright 2021, 2022, 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Module containing a base ImageData implementation suitable for use with non-DICOM
files."""

from abc import ABCMeta, abstractmethod

import numpy as np
from wsidicom import ImageData
from wsidicom.geometry import PointMm
from wsidicom.metadata import ImageCoordinateSystem


class DicomizerImageData(ImageData, metaclass=ABCMeta):
    """
    Metaclass for Dicomizer image data. Subclasses should implement all the abstract
    methods and properties in the base ImageData-class, and the pyramid_index property.
    """

    _default_z = None

    @property
    @abstractmethod
    def pyramid_index(self) -> int:
        """Should return pyramid level for image data."""
        raise NotImplementedError()

    @property
    def image_coordinate_system(self) -> ImageCoordinateSystem:
        """Return a default ImageCoordinateSystem."""
        return ImageCoordinateSystem(PointMm(0, 0), 90)

    @property
    def bits(self) -> int:
        return self.encoder.bits

    @property
    def lossy_compressed(self) -> bool:
        # TODO: This should be set from encoder and base file.
        return True

    def _encode(self, image_data: np.ndarray) -> bytes:
        """Return image data encoded in jpeg using set quality and subsample
        options.

        Parameters
        ----------
        image_data: np.ndarray
            Image data to encode.

        Returns
        ----------
        bytes
            Jpeg bytes.
        """
        return self.encoder.encode(image_data)



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Source using bioformats."""

from pathlib import Path
from typing import List, Optional, Tuple

from wsidicom.codec import Encoder
from wsidicom.metadata.wsi import WsiMetadata

from wsidicomizer.dicomizer_source import DicomizerSource
from wsidicomizer.extras.bioformats.bioformats_image_data import BioformatsImageData
from wsidicomizer.extras.bioformats.bioformats_reader import BioformatsReader
from wsidicomizer.image_data import DicomizerImageData
from wsidicomizer.metadata import WsiDicomizerMetadata


class BioformatsSource(DicomizerSource):
    def __init__(
        self,
        filepath: Path,
        encoder: Encoder,
        tile_size: Optional[int] = None,
        metadata: Optional[WsiMetadata] = None,
        default_metadata: Optional[WsiMetadata] = None,
        include_confidential: bool = True,
        readers: Optional[int] = None,
        cache_path: Optional[str] = None,
    ) -> None:
        if tile_size is None:
            raise ValueError("Tile size required for bioformats")
        self._reader = BioformatsReader(Path(filepath), readers, cache_path)
        (
            self._pyramid_image_index,
            self._label_image_index,
            self._overview_image_index,
        ) = self._get_image_indices(self._reader)
        super().__init__(
            filepath,
            encoder,
            tile_size,
            metadata,
            default_metadata,
            include_confidential,
        )

    @property
    def has_label(self) -> bool:
        return self._label_image_index is not None

    @property
    def has_overview(self) -> bool:
        return self._overview_image_index is not None

    @staticmethod
    def is_supported(filepath: Path) -> bool:
        """Return True if file in filepath is supported by Bio-Formats."""
        return BioformatsImageData.detect_format(filepath)

    @property
    def pyramid_levels(self) -> List[int]:
        """Return pyramid levels (scalings) for file."""
        return list(range(self._reader.resolution_count(self._pyramid_image_index)))

    @property
    def base_metadata(self) -> WsiDicomizerMetadata:
        return WsiDicomizerMetadata()

    @staticmethod
    def _get_image_indices(
        reader: BioformatsReader,
    ) -> Tuple[int, Optional[int], Optional[int]]:
        image_indices = list(range(reader.images_count))
        overview_image_index = None
        label_image_index = None

        for image_index in image_indices.copy():
            image_name = reader.image_name(image_index)
            if image_name is None:
                continue
            if "macro" in image_name.lower() or "overview" in image_name.lower():
                overview_image_index = image_index
                image_indices.remove(image_index)
            elif "label" in image_name.lower():
                label_image_index = image_index
                image_indices.remove(image_index)

        pyramid_image_index = 0
        largest_image_width = None
        for image_index in image_indices:
            image_width = reader.size(image_index).width
            if largest_image_width is None or largest_image_width < image_width:
                pyramid_image_index = image_index
                largest_image_width = image_width
        return pyramid_image_index, label_image_index, overview_image_index

    def _create_level_image_data(self, level_index: int) -> DicomizerImageData:
        return BioformatsImageData(
            self._reader,
            self._tile_size,
            self._encoder,
            self._pyramid_image_index,
            level_index,
        )

    def _create_label_image_data(self) -> DicomizerImageData:
        assert self._label_image_index is not None
        return BioformatsImageData(
            self._reader, self._tile_size, self._encoder, self._label_image_index, 0
        )

    def _create_overview_image_data(self) -> DicomizerImageData:
        assert self._overview_image_index is not None
        return BioformatsImageData(
            self._reader, self._tile_size, self._encoder, self._overview_image_index, 0
        )

    def close(self) -> None:
        return self._reader.close()



#    Copyright 2021, 2022, 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Source for reading czi file."""

from pathlib import Path
from typing import List, Optional

from czifile import CziFile
from wsidicom.codec import Encoder
from wsidicom.metadata import WsiMetadata

from wsidicomizer.dicomizer_source import DicomizerSource
from wsidicomizer.image_data import DicomizerImageData
from wsidicomizer.sources.czi.czi_image_data import CziImageData
from wsidicomizer.sources.czi.czi_metadata import CziMetadata


class CziSource(DicomizerSource):
    def __init__(
        self,
        filepath: Path,
        encoder: Encoder,
        tile_size: int = 512,
        metadata: Optional[WsiMetadata] = None,
        default_metadata: Optional[WsiMetadata] = None,
        include_confidential: bool = True,
    ) -> None:
        super().__init__(
            filepath,
            encoder,
            tile_size,
            metadata,
            default_metadata,
            include_confidential,
        )
        self._czi = CziFile(filepath)
        self._base_metadata = CziMetadata(self._czi)

    def close(self) -> None:
        return self._czi.close()

    @property
    def has_label(self) -> bool:
        return False

    @property
    def has_overview(self) -> bool:
        return False

    @property
    def pyramid_levels(self) -> List[int]:
        return [0]

    @property
    def base_metadata(self) -> CziMetadata:
        return self._base_metadata

    @staticmethod
    def is_supported(filepath: Path) -> bool:
        """Return True if file in filepath is supported by CziFile."""
        return CziImageData.detect_format(filepath) is not None

    def _create_level_image_data(self, level_index: int) -> DicomizerImageData:
        if level_index != 0:
            raise NotImplementedError("Only base level is supported.")
        return CziImageData(
            self._czi,
            self._tile_size,
            self._encoder,
            self.base_metadata,
            self.metadata.image,
        )

    def _create_label_image_data(self) -> DicomizerImageData:
        raise NotImplementedError("Czi does not support labels.")

    def _create_overview_image_data(self) -> DicomizerImageData:
        raise NotImplementedError("Czi does not support overviews.")



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Cli for bioformats source."""

from pathlib import Path
from typing import Optional, Sequence

from wsidicom.codec import Settings
from wsidicom.metadata.wsi import WsiMetadata

from wsidicomizer.cli import WsiDicomizerCli
from wsidicomizer.extras.bioformats.bioformats_dicomizer import BioformatsDicomizer


class BioformatsCli(WsiDicomizerCli):
    def convert(
        self,
        filepath: str,
        output_path: str,
        metadata: Optional[WsiMetadata] = None,
        default_metadata: Optional[WsiMetadata] = None,
        tile_size: int = 512,
        add_missing_levels: bool = False,
        include_levels: Optional[Sequence[int]] = None,
        include_label: bool = True,
        include_overview: bool = True,
        include_confidential: bool = True,
        workers: Optional[int] = None,
        chunk_size: Optional[int] = None,
        encoding_settings: Optional[Settings] = None,
        offset_table: str = "bot",
        label: Optional[Path] = None,
    ):
        with BioformatsDicomizer.open(
            filepath=filepath,
            metadata=metadata,
            default_metadata=default_metadata,
            tile_size=tile_size,
            include_confidential=include_confidential,
            encoding_settings=encoding_settings,
        ) as wsi:
            wsi.save(
                output_path=output_path,
                offset_table=offset_table,
                workers=workers,
                chunk_size=chunk_size,
                add_missing_levels=add_missing_levels,
                include_levels=include_levels,
                include_labels=include_label,
                include_overviews=include_overview,
                label=label,
            )


def main():
    cli = BioformatsCli()
    cli.cli()


if __name__ == "__main__":
    main()



#    Copyright 2021, 2022, 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Image data for opentile compatible file."""

from typing import Iterable, Iterator, List, Optional

from opentile.tiff_image import TiffImage
from PIL import Image as Pillow
from PIL.Image import Image
from pydicom.uid import JPEG2000, UID, JPEG2000Lossless, JPEGBaseline8Bit
from tifffile.tifffile import COMPRESSION, PHOTOMETRIC
from wsidicom.codec import Encoder
from wsidicom.geometry import Point, Size, SizeMm
from wsidicom.metadata import Image as ImageMetadata
from wsidicom.metadata import ImageCoordinateSystem

from wsidicomizer.image_data import DicomizerImageData


class OpenTileImageData(DicomizerImageData):
    def __init__(
        self,
        tiff_image: TiffImage,
        encoder: Encoder,
        force_transcoding: bool = False,
    ):
        """Wraps a TiffImage to ImageData.

        Parameters
        ----------
        tiff_image: TiffImage
            TiffImage to wrap.
        encoded: Encoder
            Encoder to use.
        force_transcoding: bool
            Force transcoding of image data.
        """
        super().__init__(encoder)
        self._tiff_image = tiff_image

        self._needs_transcoding = (
            not self.is_supported_transfer_syntax() or force_transcoding
        )
        if self.needs_transcoding:
            self._transfer_syntax = self.encoder.transfer_syntax
        else:
            self._transfer_syntax = self.get_transfer_syntax()
        self._image_size = Size(*self._tiff_image.image_size.to_tuple())
        self._tile_size = Size(*self._tiff_image.tile_size.to_tuple())
        self._tiled_size = Size(*self._tiff_image.tiled_size.to_tuple())

    def __str__(self) -> str:
        return f"{type(self).__name__} for page {self._tiff_image}"

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self._tiff_image})"

    @property
    def transfer_syntax(self) -> UID:
        """The uid of the transfer syntax of the image."""
        return self._transfer_syntax

    @property
    def needs_transcoding(self) -> bool:
        """Return true if image data requires transcoding for Dicom
        compatibility."""
        return self._needs_transcoding

    @property
    def native_compression(self) -> COMPRESSION:
        """Return compression method used in image data."""
        return self._tiff_image.compression

    @property
    def image_size(self) -> Size:
        """The pixel size of the image."""
        return self._image_size

    @property
    def tile_size(self) -> Size:
        """The pixel tile size of the image."""
        return self._tile_size

    @property
    def focal_planes(self) -> List[float]:
        """Focal planes available in the image defined in um."""
        return [self._tiff_image.focal_plane]

    @property
    def optical_paths(self) -> List[str]:
        """Optical paths available in the image."""
        return [self._tiff_image.optical_path]

    @property
    def suggested_minimum_chunk_size(self) -> int:
        """Return suggested minimum chunk size for optimal performance with
        get_encoeded_tiles()."""
        return self._tiff_image.suggested_minimum_chunk_size

    @property
    def pyramid_index(self) -> int:
        """The pyramidal index in relation to the base layer."""
        return self._tiff_image.pyramid_index

    @property
    def photometric_interpretation(self) -> str:
        if self.needs_transcoding:
            return self.encoder.photometric_interpretation
        if self._tiff_image.photometric_interpretation == PHOTOMETRIC.YCBCR:
            if self.transfer_syntax == JPEGBaseline8Bit:
                return "YBR_FULL_422"
            elif self.transfer_syntax == JPEG2000:
                return "YBR_ICT"
            elif self.transfer_syntax == JPEG2000Lossless:
                return "YBR_RCT"
        elif self._tiff_image.photometric_interpretation == PHOTOMETRIC.RGB:
            return "RGB"
        elif self._tiff_image.photometric_interpretation == (PHOTOMETRIC.MINISBLACK):
            return "MONOCHROME2"
        raise NotImplementedError(
            "Non-implemented photometric interpretation. ",
            self._tiff_image.photometric_interpretation,
        )

    @property
    def samples_per_pixel(self) -> int:
        return self._tiff_image.samples_per_pixel

    def _get_encoded_tile(self, tile: Point, z: float, path: str) -> bytes:
        """Return image bytes for tile. Returns transcoded tile if
        non-supported encoding.

        Parameters
        ----------
        tile: Point
            Tile position to get.
        z: float
            Focal plane of tile to get.
        path: str
            Optical path of tile to get.

        Returns
        ----------
        bytes
            Tile bytes.
        """
        if z not in self.focal_planes or path not in self.optical_paths:
            raise ValueError("Requested focal plane or optical path not available.")
        if self.needs_transcoding:
            decoded_tile = self._tiff_image.get_decoded_tile(tile.to_tuple())
            return self.encoder.encode(decoded_tile)
        return self._tiff_image.get_tile(tile.to_tuple())

    def _get_decoded_tile(self, tile: Point, z: float, path: str) -> Image:
        """Return Image for tile.

        Parameters
        ----------
        tile: Point
            Tile position to get.
        z: float
            Focal plane of tile to get.
        path: str
            Optical path of tile to get.

        Returns
        ----------
        Image.Image
            Tile as Image.
        """
        if z not in self.focal_planes or path not in self.optical_paths:
            raise ValueError
        return Pillow.fromarray(self._tiff_image.get_decoded_tile(tile.to_tuple()))

    def _get_encoded_tiles(
        self, tiles: Iterable[Point], z: float, path: str
    ) -> Iterator[bytes]:
        if z not in self.focal_planes or path not in self.optical_paths:
            raise ValueError
        tiles_tuples = [tile.to_tuple() for tile in tiles]
        if not self.needs_transcoding:
            return self._tiff_image.get_tiles(tiles_tuples)
        decoded_tiles = self._tiff_image.get_decoded_tiles(tiles_tuples)
        return (self.encoder.encode(tile) for tile in decoded_tiles)

    def close(self) -> None:
        self._tiff_image.close()

    def is_supported_transfer_syntax(self) -> bool:
        """Return true if image data is encoded with Dicom-supported transfer
        syntax."""
        try:
            self.get_transfer_syntax()
            return True
        except NotImplementedError:
            return False

    def get_transfer_syntax(self) -> UID:
        """Return transfer syntax (Uid) for compression type in image data."""
        compression = self.native_compression
        if compression == COMPRESSION.JPEG:
            return JPEGBaseline8Bit
        elif compression == COMPRESSION.APERIO_JP2000_RGB:
            return JPEG2000
        raise NotImplementedError(f"Not supported compression {compression}")


class OpenTileLevelImageData(OpenTileImageData):
    def __init__(
        self,
        tiff_image: TiffImage,
        image_metadata: ImageMetadata,
        merged_metadata: ImageMetadata,
        encoder: Encoder,
        force_transcoding: bool = False,
    ):
        super().__init__(tiff_image, encoder, force_transcoding)
        if (
            merged_metadata.pixel_spacing is not None
            and merged_metadata.pixel_spacing != image_metadata.pixel_spacing
        ):
            # Override pixel spacing
            self._pixel_spacing = merged_metadata.pixel_spacing * int(
                2**self._tiff_image.pyramid_index
            )
        elif self._tiff_image.pixel_spacing is not None:
            self._pixel_spacing = SizeMm(*self._tiff_image.pixel_spacing.to_tuple())
        else:
            raise ValueError("Could not determine pixel spacing for tiff image.")
        self._image_coordinate_system = merged_metadata.image_coordinate_system

    @property
    def image_coordinate_system(self) -> ImageCoordinateSystem:
        if self._image_coordinate_system is None:
            return super().image_coordinate_system
        return self._image_coordinate_system

    @property
    def pixel_spacing(self) -> SizeMm:
        return self._pixel_spacing


class OpenTileAssociatedImageData(OpenTileImageData):
    def __init__(
        self,
        tiff_image: TiffImage,
        encoder: Encoder,
        force_transcoding: bool = False,
    ):
        super().__init__(tiff_image, encoder, force_transcoding)
        if self._tiff_image.pixel_spacing is not None:
            self._pixel_spacing = SizeMm(*self._tiff_image.pixel_spacing.to_tuple())
        else:
            self._pixel_spacing = None

    @property
    def pixel_spacing(self) -> Optional[SizeMm]:
        """Size of the pixels in mm/pixel."""
        return self._pixel_spacing



#    Copyright 2021, 2022, 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Importing openslide module after loading library if needed."""

import os
from ctypes.util import find_library
from pathlib import Path
from typing import Optional

# On windows, use find_library to find directory with openslide dll in
# the Path environmental variable.
openslide_libs = ["libopenslide-1", "libopenslide-0"]
if os.name == "nt":

    def find_openslide_lib(openslide_lib_name: str) -> Optional[Path]:
        openslide_lib_path = find_library(openslide_lib_name)
        if openslide_lib_path is not None:
            return Path(openslide_lib_path)
        openslide_lib_dir = os.environ.get("OPENSLIDE")
        if openslide_lib_dir is not None:
            openslide_lib_path = (
                Path(openslide_lib_dir).joinpath(openslide_lib_name).with_suffix(".dll")
            )
            if openslide_lib_path.exists():
                return openslide_lib_path

        return None

    openslide_lib_paths = [
        path
        for path in (
            find_openslide_lib(openslide_lib) for openslide_lib in openslide_libs
        )
        if path is not None
    ]
    if len(openslide_lib_paths) == 0:
        raise ModuleNotFoundError(
            "Could not find libopenslide-0.dll or libopenslide-1.dll in the directories"
            "defined in the 'Path' and 'OPENSLIDE' environmental variables. Please add "
            "the directory with openslide bin content to the 'Path' or 'OPENSLIDE' "
            "environmental variable."
        )
    for openslide_lib_path in openslide_lib_paths:
        openslide_dir = str(openslide_lib_path.parent)
        os.add_dll_directory(openslide_dir)

from openslide import (  # noqa: E402
    PROPERTY_NAME_BACKGROUND_COLOR,  # noqa: E402
    PROPERTY_NAME_BOUNDS_HEIGHT,  # noqa: E402
    PROPERTY_NAME_BOUNDS_WIDTH,  # noqa: E402
    PROPERTY_NAME_BOUNDS_X,  # noqa: E402
    PROPERTY_NAME_BOUNDS_Y,  # noqa: E402
    PROPERTY_NAME_MPP_X,  # noqa: E402
    PROPERTY_NAME_MPP_Y,  # noqa: E402
    PROPERTY_NAME_OBJECTIVE_POWER,  # noqa: E402
    PROPERTY_NAME_VENDOR,  # noqa: E402
    OpenSlide,  # noqa: E402
)
from openslide._convert import argb2rgba as convert_argb_to_rgba  # noqa: E402
from openslide.lowlevel import _read_region, get_associated_image_names  # noqa: E402

__all__ = [
    "OpenSlide",
    "PROPERTY_NAME_BACKGROUND_COLOR",
    "PROPERTY_NAME_BOUNDS_HEIGHT",
    "PROPERTY_NAME_BOUNDS_WIDTH",
    "PROPERTY_NAME_BOUNDS_X",
    "PROPERTY_NAME_BOUNDS_Y",
    "PROPERTY_NAME_MPP_X",
    "PROPERTY_NAME_MPP_Y",
    "PROPERTY_NAME_OBJECTIVE_POWER",
    "PROPERTY_NAME_VENDOR",
    "_read_region",
    "convert_argb_to_rgba",
    "get_associated_image_names",
]



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Image data read by bioformats."""

from pathlib import Path
from typing import ContextManager, List, Optional

import numpy as np
from PIL import Image as Pillow
from PIL.Image import Image
from pydicom.uid import UID
from wsidicom.codec import Encoder
from wsidicom.geometry import Point, Region, Size, SizeMm

from wsidicomizer.extras.bioformats.bioformats_reader import BioformatsReader
from wsidicomizer.image_data import DicomizerImageData


class BioformatsImageData(DicomizerImageData):
    def __init__(
        self,
        reader: BioformatsReader,
        tile_size: int,
        encoder: Encoder,
        image_index: int,
        resolution_index: int,
    ) -> None:
        super().__init__(encoder)
        self._tile_size = Size(tile_size, tile_size)
        self._image_reader = reader
        self._image_index = image_index
        self._resolution_index = resolution_index
        self._image_region = Region(Point(0, 0), self.image_size)

    @property
    def image_region(self) -> Region:
        return self._image_region

    @property
    def pyramid_index(self) -> int:
        """Return pyramid level for image data."""
        return 0

    @property
    def files(self) -> List[Path]:
        return [Path(self._image_reader.filepath)]

    @property
    def transfer_syntax(self) -> UID:
        """Return the uid of the transfer syntax of the image."""
        return self.encoder.transfer_syntax

    @property
    def image_size(self) -> Size:
        """Return the pixel size of the image."""
        return self._image_reader.size(self._image_index, self._resolution_index)

    @property
    def tile_size(self) -> Size:
        """Return the pixel tile size of the image, or pixel size of
        the image if not tiled."""
        return self._tile_size

    @property
    def pixel_spacing(self) -> Optional[SizeMm]:
        """Return the size of the pixels in mm/pixel."""
        return self._image_reader.pixel_spacing(
            self._image_index, self._resolution_index
        )

    @property
    def samples_per_pixel(self) -> int:
        """Return number of samples per pixel (e.g. 3 for RGB)."""
        return self._image_reader.samples_per_pixel(self._image_index)

    @property
    def photometric_interpretation(self) -> str:
        """Return the photophotometric interpretation of the image
        data."""
        return self.encoder.photometric_interpretation

    def _get_tile(
        self, tile_point: Point, z: float, path: str
    ) -> ContextManager[np.ndarray]:
        region = Region(tile_point * self.tile_size, self.tile_size)
        cropped_region = self.image_region.crop(region)
        return self._image_reader.read_image(
            self._image_index, self._resolution_index, cropped_region, self.tile_size
        )

    def _get_decoded_tile(self, tile: Point, z: float, path: str) -> Image:
        """Return Image for tile defined by tile (x, y), z,
        and optical path."""
        with self._get_tile(tile, z, path) as data:
            return Pillow.fromarray(data)

    def _get_encoded_tile(self, tile: Point, z: float, path: str) -> bytes:
        """Return image bytes for tile defined by tile (x, y), z,
        and optical path."""
        with self._get_tile(tile, z, path) as data:
            return self.encoder.encode(data)

    @staticmethod
    def detect_format(filepath: Path) -> bool:
        return BioformatsReader.is_supported(filepath)



#    Copyright 2021, 2022, 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Metadata for czi file."""

from datetime import datetime
from functools import cached_property
from typing import List, Optional, Sequence, Tuple, Type, TypeVar
from xml.etree import ElementTree

import numpy as np
from czifile import CziFile
from wsidicom.geometry import SizeMm
from wsidicom.metadata import Equipment, Image, Objectives, OpticalPath

from wsidicomizer.metadata import WsiDicomizerMetadata

ElementType = TypeVar("ElementType", str, int, float)


class CziMetadata(WsiDicomizerMetadata):
    def __init__(self, czi: CziFile):
        metadata_xml = czi.metadata()
        if metadata_xml is None or not isinstance(metadata_xml, str):
            raise ValueError("No metadata string in file.")
        self._metadata = ElementTree.fromstring(metadata_xml)
        image = Image(
            acquisition_datetime=self.aquisition_datetime,
            pixel_spacing=self.pixel_spacing,
        )
        equipment = Equipment(
            model_name=self.scanner_model,
            software_versions=self.scanner_software_versions,
        )
        optical_paths = [
            OpticalPath("0", objective=Objectives(objective_power=self.magnification))
        ]
        super().__init__(equipment=equipment, image=image, optical_paths=optical_paths)

    @property
    def aquisition_datetime(self) -> Optional[datetime]:
        value = self.get_value_from_element(
            self._metadata,
            "AcquisitionDateAndTime",
            str,
            nested=["Metadata", "Information", "Image"],
        )
        return datetime.fromisoformat(value)

    @property
    def scanner_model(self) -> Optional[str]:
        information = self.get_nested_element(["Metadata", "Information"])
        image = self.get_nested_element(["Image"], information)
        microscope_ref = self.get_element(image, "MicroscopeRef").get("Id")
        microscopes = self.get_nested_element(
            ["Instrument", "Microscopes"], information
        ).findall("Microscope")
        try:
            microscope = next(
                microscope
                for microscope in microscopes
                if microscope.get("Id") == microscope_ref
            )
        except StopIteration:
            return None
        return microscope.get("Name")

    @property
    def magnification(self) -> Optional[float]:
        information = self.get_nested_element(["Metadata", "Information"])
        objective_refs = [
            objective.get("Id")
            for objective in self.get_nested_element(
                ["Image", "ObjectiveSettings"], information
            ).findall("ObjectiveRef")
        ]
        if len(objective_refs) != 1:
            return None
        objectives = self.get_nested_element(
            ["Instrument", "Objectives"], information
        ).findall("Objective")
        try:
            objective = next(
                objective
                for objective in objectives
                if objective.get("Id") == objective_refs[0]
            )
        except StopIteration:
            return None
        try:
            return self.get_value_from_element(objective, "NominalMagnification", float)
        except ValueError:
            return None

    @property
    def scanner_software_versions(self) -> Optional[List[str]]:
        application = self.get_nested_element(
            ["Metadata", "Information", "Application"]
        )
        try:
            name = self.get_value_from_element(application, "Name", str)
            version = self.get_value_from_element(application, "Version", str)
        except ValueError:
            return None
        return [name + " " + version]

    @cached_property
    def scaling(self) -> Tuple[Optional[float], Optional[float], Optional[float]]:
        scaling_elements = self.get_nested_element(["Metadata", "Scaling", "Items"])
        x: Optional[float] = None
        y: Optional[float] = None
        z: Optional[float] = None
        for distance in scaling_elements.findall("Distance"):
            dimension = distance.get("Id")
            # Value is in m per pixel, result in mm per pixel
            value = self.get_value_from_element(
                distance,
                "Value",
                float,
            ) * pow(10, 6)
            if dimension == "X":
                x = value
            elif dimension == "Y":
                y = value
            elif dimension == "Z":
                z = value
        return x, y, z

    @cached_property
    def pixel_spacing(self) -> SizeMm:
        """Get pixel spacing (mm per pixel) from metadata"""
        x, y, _ = self.scaling
        if x is None or y is None:
            raise ValueError("Could not find pixel spacing in metadata")
        return SizeMm(x, y) / 1000

    @cached_property
    def focal_plane_mapping(self) -> List[float]:
        image = self.get_nested_element(["Metadata", "Information", "Image"])
        try:
            size_z = self.get_value_from_element(image, "SizeZ", int, 0)
            z_interval = self.get_nested_element(
                ["Dimensions", "Z", "Positions", "Interval"], image
            )
            start = self.get_value_from_element(z_interval, "Start", int)
            increment = self.get_value_from_element(z_interval, "Increment", int)
            _, _, z_scale = self.scaling
            if z_scale is None:
                raise ValueError("No z scale in metadata")
            start_z = start * z_scale
            end_z = (start + increment * size_z) * z_scale
            step_z = increment * z_scale
            return list(np.arange(start_z, end_z, step_z))
        except ValueError:
            return [0.0]

    @cached_property
    def channel_mapping(self) -> List[str]:
        channels = self.get_nested_element(
            ["Metadata", "Information", "Image", "Dimensions", "Channels"]
        )
        return [
            self.get_value_from_element(channel, "Fluor", str) for channel in channels
        ]

    def get_nested_element(
        self, tags: Sequence[str], element: Optional[ElementTree.Element] = None
    ) -> ElementTree.Element:
        if element is None:
            element = self._metadata
        found_element = element
        for tag in tags:
            found_element = found_element.find(tag)
            if found_element is None:
                raise ValueError(f"Tag {tag} not found in element")
        return found_element

    def get_value_from_element(
        self,
        element: ElementTree.Element,
        tag: str,
        value_type: Type[ElementType],
        default: Optional[ElementType] = None,
        nested: Optional[Sequence[str]] = None,
    ) -> ElementType:
        if nested is not None:
            element = self.get_nested_element(nested, element)
        try:
            element = self.get_element(element, tag)
            text = element.text
            if text is None:
                raise ValueError("Text not found in element")
        except ValueError as exception:
            if default is not None:
                return default
            raise ValueError(f"Tag {tag} or text not found in element") from exception
        try:
            return value_type(text)
        except ValueError as exception:
            raise ValueError(
                f"Failed to convert tag {tag} value {text} to {value_type}."
            ) from exception

    @staticmethod
    def get_element(element: ElementTree.Element, tag: str) -> ElementTree.Element:
        found_element = element.find(tag)
        if found_element is None:
            raise ValueError(f"Tag {tag} not found in element")
        return found_element



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Dicomizer for bioformats source."""

from pathlib import Path
from typing import Optional, Union

from wsidicom import WsiDicom
from wsidicom.codec import Encoder, JpegSettings
from wsidicom.codec import Settings as EncodingSettings
from wsidicom.metadata.wsi import WsiMetadata

from wsidicomizer.extras.bioformats.bioformats_source import BioformatsSource
from wsidicomizer.wsidicomizer import WsiDicomizer


class BioformatsDicomizer(WsiDicomizer):
    @classmethod
    def open(
        cls,
        filepath: Union[str, Path],
        metadata: Optional[WsiMetadata] = None,
        default_metadata: Optional[WsiMetadata] = None,
        tile_size: int = 512,
        include_confidential: bool = True,
        encoding_settings: Optional[EncodingSettings] = None,
    ) -> WsiDicom:
        """Open data in file in filepath as WsiDicom.

        Parameters
        ----------
        filepath: str
            Path to file
        metadata: Optional[WsiMetadata] = None
            User-specified metadata that will overload metadata from source image file.
        default_metadata: Optional[WsiMetadata] = None
            User-specified metadata that will be used as default values.
        tile_size: int = 512
            Tile size to use if not defined by file.
        include_confidential: bool = True
            Include confidential metadata.
        encoding: Optional[Union[EncodingSettings, Encoder]] = None,
            Encoding setting or encoder to use if re-encoding.


        Returns
        ----------
        WsiDicom
            WsiDicom object of file.
        """
        if not isinstance(filepath, Path):
            filepath = Path(filepath)

        if not BioformatsSource.is_supported(filepath):
            raise NotImplementedError(f"{filepath} is not supported")
        if encoding_settings is None:
            encoding_settings = JpegSettings()
        encoder = Encoder.create_for_settings(encoding_settings)

        dicomizer = BioformatsSource(
            filepath,
            encoder,
            tile_size,
            metadata,
            default_metadata,
            include_confidential,
        )
        return cls(dicomizer)



#    Copyright 2021, 2022, 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""
Main module containing the WsiDicomizer class that allows non-DICOM files to be opened
like DICOM instances, enabling viewing and saving.
"""

import os
from pathlib import Path
from typing import Callable, List, Optional, Sequence, Type, Union

from PIL.Image import Image
from pydicom.uid import UID, generate_uid
from wsidicom import WsiDicom
from wsidicom.codec import Encoder, JpegSettings
from wsidicom.codec import Settings as EncodingSettings
from wsidicom.file import OffsetTableType
from wsidicom.metadata import WsiMetadata

from wsidicomizer.dicomizer_source import DicomizerSource
from wsidicomizer.sources import CziSource, OpenTileSource, TiffSlideSource

# List of supported Dicomizers in prioritization order.
loaded_sources: List[Type[DicomizerSource]] = [
    OpenTileSource,
    TiffSlideSource,
    CziSource,
]

try:
    from wsidicomizer.extras.openslide import OpenSlideSource

    loaded_sources.append(OpenSlideSource)
except ImportError:
    pass


class WsiDicomizer(WsiDicom):
    """Interface for Dicomizing files."""

    @classmethod
    def open(
        cls,
        filepath: Union[str, Path],
        metadata: Optional[WsiMetadata] = None,
        default_metadata: Optional[WsiMetadata] = None,
        tile_size: int = 512,
        include_confidential: bool = True,
        encoding: Optional[Union[EncodingSettings, Encoder]] = None,
        preferred_source: Optional[Type[DicomizerSource]] = None,
        **source_args,
    ) -> WsiDicom:
        """Open data in file in filepath as WsiDicom.

        Parameters
        ----------
        filepath: str
            Path to file
        metadata: Optional[WsiMetadata] = None
            User-specified metadata that will overload metadata from source image file.
        default_metadata: Optional[WsiMetadata] = None
            User-specified metadata that will be used as default values.
        tile_size: int = 512
            Tile size to use if not defined by file.
        include_confidential: bool = True
            Include confidential metadata.
        encoding: Optional[Union[EncodingSettings, Encoder]] = None,
            Encoding setting or encoder to use if re-encoding.
        preferred_source: Optional[Type[DicomizerSource]] = None
            Optional override source to use.
        **source_args
            Optional keyword args to pass to source.

        Returns
        ----------
        WsiDicom
            WsiDicom object of file.
        """
        if not isinstance(filepath, Path):
            filepath = Path(filepath)

        selected_source = None
        if preferred_source is None:
            selected_source = next(
                (source for source in loaded_sources if source.is_supported(filepath)),
                None,
            )
        elif preferred_source.is_supported(filepath):
            selected_source = preferred_source
        if selected_source is None:
            raise NotImplementedError(f"{filepath} is not supported")
        if encoding is None:
            encoding = JpegSettings()
        if isinstance(encoding, EncodingSettings):
            encoder = Encoder.create_for_settings(encoding)
        else:
            encoder = encoding

        source = selected_source(
            filepath,
            encoder,
            tile_size,
            metadata,
            default_metadata,
            include_confidential,
            **source_args,
        )
        return cls(source, True)

    @classmethod
    def convert(
        cls,
        filepath: Union[str, Path],
        output_path: Optional[Union[str, Path]] = None,
        metadata: Optional[WsiMetadata] = None,
        default_metadata: Optional[WsiMetadata] = None,
        tile_size: int = 512,
        uid_generator: Callable[..., UID] = generate_uid,
        add_missing_levels: bool = False,
        include_levels: Optional[Sequence[int]] = None,
        include_label: bool = True,
        include_overview: bool = True,
        include_confidential: bool = True,
        label: Optional[Union[Image, str, Path]] = None,
        workers: Optional[int] = None,
        chunk_size: Optional[int] = None,
        encoding: Optional[Union[Encoder, EncodingSettings]] = None,
        offset_table: Union["str", OffsetTableType] = OffsetTableType.BASIC,
        preferred_source: Optional[Type[DicomizerSource]] = None,
        **source_args,
    ) -> List[str]:
        """Convert data in file to DICOM files in output path. Created
        instances get UID from uid_generator. Closes when finished.

        Parameters
        ----------
        filepath: Union[str, Path],
            Path to file
        output_path: str = None
            Folder path to save files to.
        metadata: Optional[WsiMetadata] = None
            User-specified metadata that will overload metadata from source image file.
        default_metadata: Optional[WsiMetadata] = None
            User-specified metadata that will be used as default values.
        tile_size: int = 512
            Tile size to use if not defined by file.
        uid_generator: Callable[..., UID] = generate_uid
             Function that can generate unique identifiers.
        add_missing_levels: bool = False
            If to add missing dyadic levels up to the single tile level.
        include_levels: Optional[Sequence[int]] = None
            Optional list indices (in present levels) to include, e.g. [0, 1]
            includes the two lowest levels. Negative indices can be used,
            e.g. [-1, -2] includes the two highest levels.
        include_label: bool = True
            Include label(s), default true.
        include_overwiew: bool = True
            Include overview(s), default true.
        include_confidential: bool = True
            Include confidential metadata.
        label: Optional[Union[Image, str, Path]] = None,
            Optional label image to use instead of label found in file.
        workers: Optional[int] = None,
            Maximum number of thread workers to use.
        chunk_size: Optional[int] = None,
            Chunk size (number of tiles) to process at a time. Actual chunk
            size also depends on minimun_chunk_size from image_data.
        encoding: Optional[Union[EncodingSettings, Encoder]] = None,
            Encoding setting or encoder to use if re-encoding.
        offset_table: Union["str", OffsetTableType] = OffsetTableType.BASIC,
            Offset table to use, 'bot' basic offset table, 'eot' extended
            offset table, 'empty' - empty offset table.
        preferred_source: Optional[Type[DicomizerSource]] = None
            Optional override source to use.
        **source_args
            Optional keyword args to pass to source.

        Returns
        ----------
        List[str]
            List of paths of created files.
        """
        with cls.open(
            filepath,
            metadata,
            default_metadata,
            tile_size,
            include_confidential,
            encoding,
            preferred_source,
            **source_args,
        ) as wsi:
            if output_path is None:
                output_path = str(
                    Path(filepath).parents[0].joinpath(Path(filepath).stem)
                )
            try:
                os.mkdir(output_path)
            except FileExistsError:
                ValueError(f"Output path {output_path} already exists")
            created_files = wsi.save(
                output_path,
                uid_generator,
                workers,
                chunk_size,
                offset_table,
                include_levels=include_levels,
                include_labels=include_label,
                include_overviews=include_overview,
                add_missing_levels=add_missing_levels,
                label=label,
            )

        return [str(filepath) for filepath in created_files]
